from dkist_fits_specifications.spec122 import (
    define_122_schema_expansions,
    load_raw_spec122,
    load_spec122,
)
from dkist_fits_specifications.utils.frozendict import frozendict


def test_load_122():
    spec = load_spec122()
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], frozendict)
    assert isinstance(spec["fits"]["NAXIS"], frozendict)


def test_load_raw_122():
    spec = load_raw_spec122()
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], tuple)
    header, spec = spec["fits"]
    assert isinstance(header, frozendict)
    assert isinstance(spec["NAXIS"], frozendict)


def test_define_122_schema_expansion_duplication():
    """
    Given: the list of requested spec 122 expansions
    When: checking the indices for each expansion
    Then: None of them match (all expansions are unique)
    """
    expansions = define_122_schema_expansions()
    expansion_indices = [e.index for e in expansions]
    assert len(expansion_indices) == len(set(expansion_indices))



# Copyright (c) 2012 Santiago Lezica
# Used under the terms of the MIT License: https://github.com/slezica/python-frozendict

from collections.abc import Mapping

__all__ = ["frozendict"]


class frozendict(Mapping):
    """
    An immutable wrapper around dictionaries that implements the complete :py:class:`collections.Mapping`
    interface. It can be used as a drop-in replacement for dictionaries where immutability is desired.
    """

    dict_cls = dict

    def __init__(self, *args, **kwargs):
        self._dict = self.dict_cls(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def copy(self, **add_or_replace):
        return self.__class__(self, **add_or_replace)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._dict!r}>"

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in self._dict.items():
                h ^= hash((key, value))
            self._hash = h
        return self._hash



"""
This file contains sphinx directives for rendering a 122 or a 214 spec given it's filename root.
"""
from typing import Dict, List
from collections import defaultdict

from docutils import nodes, statemachine
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.tables import ListTable
from sphinx.util.nodes import nested_parse_with_titles

from dkist_fits_specifications.spec122 import load_spec122
from dkist_fits_specifications.spec214 import load_full_spec214, load_raw_spec214


class BaseSpecTable(ListTable):
    """
    Given a pre-loaded representation of a spec table render it.
    """

    option_spec = {"header-level": directives.single_char_or_unicode}

    # Stolen from https://github.com/sphinx-doc/sphinx/issues/8039
    # A method which takes an arbitary rst string and converts it to Nodes
    def convert_rst_to_nodes(self, rst_source: str) -> List[nodes.Node]:
        """Turn an RST string into a node that can be used in the document."""
        node = nodes.Element()
        node.document = self.state.document
        nested_parse_with_titles(
            state=self.state,
            content=statemachine.ViewList(
                statemachine.string2lines(rst_source),
                # The following line isn't used but it sometimes reported in error messages
                source=f"[Custom rst preamble for {self.arguments[0]} table]",
            ),
            node=node,
        )
        return node.children

    def build_table_data(self, spec_table, colnames) -> List[List[nodes.Node]]:
        """
        Generate a representation of the table in a format compatible with the input to ListTable.build_table_from_list.

        This means that all the text elements need to be manually parsed into
        docutils "nodes" which are the intermediate format between rst and html
        (or other output) in docutils.

        The output is a list of rows, each element in a row being a ``Node`` to be rendered in that cell.
        """
        table_rows = [
            [nodes.paragraph(text="Key"), *[nodes.paragraph(text=col) for col in colnames.values()]]
        ]
        for key, info in spec_table.items():
            row = [nodes.literal(text=key)]
            for col in colnames.keys():
                value = info.get(col, "")
                if isinstance(value, list):
                    node = nodes.bullet_list(bullet="*")
                    # Add a css class to this list so we can handle it in the theme
                    node["classes"].append("spec-table-ul")
                    for item in value:
                        li = nodes.list_item()
                        li.append(nodes.paragraph(text=item))
                        node.append(li)
                else:
                    node = nodes.paragraph(text=value)
                row.append(node)
            table_rows.append(row)

        return table_rows

    def get_list_table_extra(self, table_data):
        """
        Generate the keyword arguments for ListTable.build_table_from_list.

        In a method so it can be overriden.
        """
        num_cols = len(table_data[0])
        col_widths = self.get_column_widths(num_cols)
        header_rows = self.options.get("header-rows", 1)
        stub_columns = self.options.get("stub-columns", 0)
        return {
            "col_widths": col_widths,
            "header_rows": header_rows,
            "stub_columns": stub_columns,
        }

    def generate_table_node(self, table_data):
        """
        Convert the table_data format into a ``nodes.table``.
        """
        table_node = self.build_table_from_list(table_data, **self.get_list_table_extra(table_data))

        if "colwidths-given" in table_node:
            table_node["classes"].remove("colwidths-given")
        if "colwidths-auto" not in table_node:
            table_node["classes"] += ["colwidths-auto"]

        return table_node

    def run(self):
        """
        This method is the entry point for the Directive.

        As much as possible is done here by calling other methods so the 214
        and 122 subclasses can override the individual components they need.
        """
        preamble = self.get_preamble()
        preamble_nodes = self.convert_rst_to_nodes(preamble) if preamble else []
        spec_table = self.get_spec_table()
        spec_table = {key: dict(spec) for key, spec in spec_table.items()}
        spec_table = self.prep_table(spec_table)
        table_data = self.build_table_data(spec_table, self.get_colnames(spec_table))
        table_node = self.generate_table_node(table_data)
        return preamble_nodes + [table_node]

    def get_colnames(self, spec_table):
        """
        Get the column names to be displayed.

        By default this is any field for each key in the header.
        Override this method to add custom filtering of the keys.
        """
        all_keys = list(
            set.union(*[set(keys) for keys in [info.keys() for info in spec_table.values()]])
        )
        return dict((zip(all_keys, all_keys)))

    def get_preamble(self):
        """
        Get a rst block which will be placed before the table.
        """
        return ""

    def get_spec_table(self) -> Dict[str, Dict[str, str]]:
        """
        Get the spec_table as defined in `dkist_fits_specifications`.

        The format of the return is a dict with a key matching the FITS header
        key and the value a dict containing all the data about that key.

        Each element of this dict is converted to a row in the table.
        """
        raise NotImplementedError("Subclasses must implement get_spec_table")

    def prep_table(self, spec_table):
        """
        Perform any desired pre-processing on the spec_table before converting it into a table.
        """
        return spec_table


class Spec214Table(BaseSpecTable):
    """
    Generate a spec table from the 214 specs.
    """

    required_arguments = 1

    def get_colnames(self, spec_table):
        key_map = {
            "type": "Type",
            "description": "Description",
            "fitsreference": "Reference to FITS 4.0",
            "units": "Unit",
            "values": "Allowed Values",
        }

        # If a column is empty then don't use it.
        columns = defaultdict(list)
        for key, info in spec_table.items():
            for col, value in info.items():
                columns[col].append(value)
        for col in list(key_map.keys()):
            if all([not v for v in columns[col]]):
                key_map.pop(col)
        return key_map

    def get_spec_table(self):
        spec_table = load_full_spec214(self.arguments[0])
        return spec_table[self.arguments[0]]

    def get_preamble(self):
        spec_header = load_raw_spec214(self.arguments[0])
        spec_header = spec_header[self.arguments[0]][0]["spec214"]
        title = spec_header["title"] + " Keywords"
        header_level = self.options.get("header-level", "#")
        lines = [
            title,
            header_level * len(title),
            "",
            spec_header.get("summary", ""),
            "",
            spec_header.get("long_description", ""),
        ]
        return "\n".join(lines)

    def prep_table(self, spec_table):
        type_lookup = {
            "str": "string",
            "bool": "boolean",
            "int": "integer",
            "float": "float",
        }

        for key, info in spec_table.items():
            if info["units"] is None:
                del info["units"]
            info["type"] = type_lookup[info["type"]]
            if info.get("format"):
                info["format"] = "time" if info["format"] == "isot" else info["format"]
                info["type"] = f'{info["type"]} ({info["format"]})'
        return spec_table


class Spec122Table(BaseSpecTable):
    """
    Generate a spec table from the 122 specs.
    """

    required_arguments = 1

    def get_colnames(self, spec_table):
        return {
            "type": "Type",
            "comment": "FITS Comment",
            "required": "Required by the Data center",
            "expected": "Expected to always be present",
        }

    def get_spec_table(self):
        spec_table = load_spec122(self.arguments[0])
        return spec_table[self.arguments[0]]


def setup(app):
    app.add_directive("spec-214-table", Spec214Table)
    app.add_directive("spec-122-table", Spec122Table)

    return {"parallel_read_safe": True, "parallel_write_safe": True}



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.

# -- Project information -----------------------------------------------------

project = "dkist-fits-specifications"

# The full version, including alpha/beta/rc tags
from dkist_fits_specifications import __version__

release = __version__

# -- General configuration ---------------------------------------------------

from dkist_sphinx_theme.conf.core import *

extensions += [
    "dkist_fits_specifications.utils.sphinx.spec_table",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
]
extensions.remove("autoapi.extension")

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

from dkist_sphinx_theme.conf.theme import *



"""
Functions and schemas relating to specification 122, the level 0 FITS files.

The 122 schemas have a singular variant, they are all NAXIS=3.
However, the yaml files still are written with the indices corresponding to NAXIS (``n, i, j``).
These indices are expanded by the loading function so the returned schema always has ``NAXIS = 3``.

The yamls are written with the indices, to reduce duplication and to make it
easier for 214 to use the raw (unprocessed) schemas.
"""
import copy
from typing import Optional
from pathlib import Path
from functools import cache

import yamale  # type: ignore

from dkist_fits_specifications import validation_schema
from dkist_fits_specifications.utils import (
    frozendict,
    load_raw_spec,
    raw_schema_type_hint,
    schema_type_hint,
)
from dkist_fits_specifications.utils.spec_processors import (
    ExpansionIndex,
    expand_schema,
)

__all__ = ["load_raw_spec122", "load_spec122"]


base_path = Path(__file__).parent / "schemas"


def define_122_schema_expansions() -> list[ExpansionIndex]:
    # 122 always has 3 axes, but we encode the yamls independent of the number,
    # to match 214 which might have a variable number of axes.
    n_expansion = ExpansionIndex(index="n", size=1, values=range(1, 4))
    i_expansion = ExpansionIndex(index="i", size=1, values=range(1, 4))
    j_expansion = ExpansionIndex(index="j", size=1, values=range(1, 4))
    return [n_expansion, i_expansion, j_expansion]


def preprocess_schema(schema: schema_type_hint) -> schema_type_hint:
    """
    Convert the loaded raw schemas to the 122 schema.

    Parameters
    ----------
    raw_schema
        The loaded version of a single yaml file.

    Returns
    -------
    schema
        The body of a schema, updated as needed from the yaml files.
    """
    header, raw_schemas = schema
    header = dict(copy.deepcopy(header)["spec122"])
    header.pop("section")

    schema = {
        k: v
        for k, v in expand_schema(
            schema=raw_schemas, expansions=define_122_schema_expansions()
        ).items()
    }
    for key, key_schema in schema.items():
        updated_schema = {key: {**header, **key_schema}}
        # Rather than put expected in all the files, default it to required
        updated_schema[key]["expected"] = key_schema.get("expected", key_schema["required"])
        updated_schema[key] = frozendict(updated_schema[key])
        schema.update(updated_schema)
    return frozendict(schema)


# No cache here as load_raw_spec is cached
def load_raw_spec122(glob: Optional[str] = None) -> frozendict[str, raw_schema_type_hint]:
    """
    Load the raw 122 schemas from the yaml files.

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    Returns
    -------
    raw_schemas
        The schemas as loaded from the yaml files.
    """
    return load_raw_spec(base_path, glob)


@cache
def load_spec122(glob: Optional[str] = None) -> frozendict[str, schema_type_hint]:
    """
    Return the loaded schemas for DKIST Specification 122

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.
    """

    raw_schemas = load_raw_spec122(glob=glob)
    schemas = {}
    for schema_name, raw_schema in raw_schemas.items():
        # 122 only uses the second document
        schema = preprocess_schema(raw_schema)
        yamale.validate(validation_schema, [(schema, None)])
        schemas[schema_name] = schema

    return frozendict(schemas)



"""
DKIST FITS header specifications.

This package contains machine readable versions of the DKIST specifications 122 and 214.
The objectives of this package are to:

* Provide a single source of truth for all software interacting with DKIST FITS specifications.
* Reduce duplication in the specifications themselves, to ease reading and updating the specifications.
* Provide a coupling between the level 1 (214) specification and the level 0
  (122) specification, as much of the metadata is copied from the level 0 to
  the level 1 headers, so if the 122 specification changes the 214
  specification is automatically updated to match.

"""
from pathlib import Path

from yamale import make_schema  # type: ignore

from .version import __version__  # type: ignore

__all__ = []


validation_schema = make_schema(Path(__file__).parent / "spec_validation_schema.yml")



import logging
from abc import ABC, abstractmethod
from typing import Any, Iterable
from dataclasses import dataclass

logger = logging.getLogger(__name__)
from dkist_fits_specifications.utils import frozendict, schema_type_hint

# Defined as a random string so we don't accidentally confuse ourselves in tests where we want to
# use a fake instrument name
_NOT_A_DKIST_INSTRUMENT = "671976988c6c4bf78836401062757b16"


class ConditionalRequirement(ABC):
    """
    A class definition for capturing a single type of conditional requiredness.

    These classes are instantiated with a header and then can be used to check the requiredness of any key in the schema
    """

    def __init__(self, header: dict[str:Any]):
        self.header = header

    @abstractmethod
    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Given the schema for a single key, check if this key needs to be required."""
        pass


class InstrumentRequired(ConditionalRequirement):
    """
    Conditional requirement that makes all keys related to the headers "INSTRUME" required.

    spec_field:
      instrument_required: str
    """

    def __init__(self, header: dict[str, Any]):
        super().__init__(header=header)
        self.instrument = self.header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()

    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Check if the required instrument matches the instrument in the header."""
        # "False" is importantly different than _NOT_A_DKIST_INSTRUMENT
        return spec_fields.get("instrument_required", "False").casefold() == self.instrument


def _vbi_is_polarimetric(header: dict[str, Any]) -> False:
    """VBI is *always* non-polarimetric."""
    return False


def _visp_is_polarimetric(header: dict[str, Any]) -> bool:
    """Check ViSP polarimeter mode to see if dataset was polarimetric."""
    return header["VSPPOLMD"] == "observe_polarimetric"


def _cryonirsp_is_polarimetric(header: dict[str, Any]) -> bool:
    """
    Check Cryo-NIRSP headers to see if dataset was polarimetric.

    Polarimetric if there are more than 1 modstate and the modulator was in a "moving" mode.
    """
    return header["CNMODNST"] > 1 and header["CNSPINMD"] in ["Continuous", "Stepped"]


def _dlnirsp_is_polarimetric(header: dict[str, Any]) -> bool:
    """Check DL-NIRSP polarimeter mode to see if dataset was polarimetric."""
    return header["DLPOLMD"] == "Full Stokes"


def _fallback_is_polarimetric(header: dict[str, Any]) -> False:
    """
    One-stop function for when we can't figure out what to do with a particular instrument.

    Logs a warning and returns False.
    """
    instrument = header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()
    logger.warning(
        f"Checking polarimetric requiredness for instrument {instrument} is not currently supported."
    )
    return False


class PolarimetricRequired(ConditionalRequirement):

    """
    Conditional requirement that makes keys required if the dataset represents polarimetric data.

    spec_field:
      polarimetric_required: bool
    """

    instrument_to_polarimetric_func_mapping = {
        "vbi": _vbi_is_polarimetric,
        "visp": _visp_is_polarimetric,
        "cryo-nirsp": _cryonirsp_is_polarimetric,
        "dl-nirsp": _dlnirsp_is_polarimetric,
    }

    def __init__(self, header: dict[str, Any]):
        super().__init__(header=header)
        instrument = self.header.get("INSTRUME", _NOT_A_DKIST_INSTRUMENT).casefold()

        is_polarimetric_function = self.instrument_to_polarimetric_func_mapping.get(
            instrument, _fallback_is_polarimetric
        )

        try:
            self.is_polarimetric = is_polarimetric_function(self.header)
        except Exception as e:
            logger.warning(
                f"Error encountered when checking polarimetric requiredness. Error was:\n{e}"
            )
            self.is_polarimetric = False

    def check_requiredness(self, spec_fields: frozendict[str, Any]) -> bool:
        """Check if the header was from a polarimetric dataset and the key is `polarimetric_required`."""
        return self.is_polarimetric and spec_fields.get("polarimetric_required", False)


def update_schema_requiredness(
    schema: schema_type_hint, requirements: list[ConditionalRequirement]
) -> dict[str, frozendict[str, Any]]:
    """
    Modify a schema's `required` values based on conditional requirements.

    If ANY of the conditional requiredness requirements are met then the key is set to "required."
    """
    updated_schema = dict()
    for fits_keyword_name, spec_fields in schema.items():
        thawed_fields = dict(spec_fields)

        if any([requirement.check_requiredness(spec_fields) for requirement in requirements]):
            thawed_fields["required"] = True

        updated_schema[fits_keyword_name] = frozendict(thawed_fields)

    return updated_schema


@dataclass
class ExpansionIndex:
    """
    A class for defining a FITS schema expansion.

    index: the string to be substituted, omitting the surrounding '<', '>'
    size: how many (zero-padded) characters to use for the substituted integer
    values: the iterable of integers to be used as substitution for the index

    Example:
        Using the expansion of:
        ExpansionIndex(index="a", size=3, values=range(1, 6))

        on the keyword 'KEY<a>' would produce the expanded set of keys:
        ['KEY001', 'KEY002', 'KEY003', 'KEY004', 'KEY005']

    """

    index: str
    values: Iterable
    size: int = None

    def __post_init__(self):
        if len(str(max(self.values))) > self.size:
            raise ValueError(
                f"The maximum expansion value ({max(self.values)}) does not fit within the prescribed size ({self.size})."
            )

    def _expanded_keys(self, key: str) -> list[str]:
        """Generate schema entries for expanded keys."""
        return [key.replace(f"<{self.index}>", str(i).zfill(self.size)) for i in self.values]

    def generate(self, keys: list[str]) -> list[str]:
        """Generate the new keys to be added."""
        return_keys = []
        for key in keys:
            if f"<{self.index}>" in key:
                return_keys.extend(self._expanded_keys(key=key))
        long_keys = [k for k in return_keys if len(k) > 8]
        if long_keys:
            raise ValueError(
                f"FITS keywords cannot be more than 8 characters in length. {long_keys} are too long."
            )
        return return_keys


def expand_schema(
    schema: schema_type_hint, expansions: list[ExpansionIndex]
) -> dict[str, frozendict[str, Any]]:
    """Perform a schema expansion given a schema and a list of ExpansionIndexes to apply."""
    expanded_schema = dict()
    for fits_keyword_name, spec_fields in schema.items():
        if "<" not in fits_keyword_name:
            expanded_schema.update({fits_keyword_name: spec_fields})
        else:
            expanded_fits_keywords = [fits_keyword_name]
            for expansion in expansions:
                expanded_fits_keywords.extend(expansion.generate(keys=expanded_fits_keywords))
            expanded_schema.update({k: spec_fields for k in expanded_fits_keywords if "<" not in k})
    return expanded_schema



"""
Functions for formatting FITS headers.
"""
import textwrap
from functools import partial

from astropy.io import fits

from dkist_fits_specifications.spec214 import (
    expand_214_schema,
    load_full_spec214,
    load_raw_spec214,
)

__all__ = ["reformat_spec214_header", "HEADER_SECTION_ORDER_214"]


HEADER_SECTION_ORDER_214 = [
    "fits",
    "telescope",
    "datacenter",
    "dataset",
    "stats",
    "dkist-id",
    "dkist-op",
    "camera",
    "pac",
    "ao",
    "wfc",
    "ws",
    "vbi",
    "visp",
    "cryonirsp",
    "dlnirsp",
    "vtf",
    "compression",
]


def _get_comment(key, spec, input_header):
    """
    Get the comment from the various places the comment could be.
    """
    comment = input_header.comments[key] if key in input_header.comments else None
    return comment or spec[key].get("comment") or None


def symmetrically_pad(string, width=72, filler="-"):
    """
    Pad a string with filler on both the left and right.
    """
    # If we have more than one line, then wrap the lines and pad each individual line
    if len(string) > width:
        # subtract from width 2 to always have one filler at each end
        lines = textwrap.wrap(string, width=width - 2)
        return "".join(map(partial(symmetrically_pad, width=width, filler=filler), lines))

    string = f" {string} "
    left_pad = int((width - len(string)) / 2)
    string = f"{string :{filler}>{left_pad + len(string)}}"
    string = f"{string :{filler}<{width}}"
    return string


def reformat_spec214_header(input_header):
    """
    Using the information in the yamls reformat a FITS header.

    This adds title and summary fields in the yamls as comment blocks between
    the sections as well as ordering the keys according to their order in the
    specifications.
    """
    spec214 = load_full_spec214()
    output_header = fits.Header()
    for section in HEADER_SECTION_ORDER_214:
        spec = spec214[section]
        spec = expand_214_schema(spec, **dict(input_header))
        spec_preamble = load_raw_spec214(section)[section][0]["spec214"]

        # Skip a section if we have no keys for that section
        if not any([key in input_header for key in spec.keys()]):
            continue

        # Don't mess around with default FITS keys
        if section not in ["fits", "compression"]:
            output_header.add_blank(after=-1)
            output_header.add_comment(symmetrically_pad(spec_preamble["title"]), after=-1)
            if "summary" in spec_preamble:
                output_header.add_comment(
                    symmetrically_pad(spec_preamble["summary"], filler=" "), after=-1
                )
            output_header.add_comment("-" * 72, after=-1)

        for key in spec.keys():
            if key not in input_header or key in output_header:
                continue
            output_header.append(
                fits.Card(
                    key=key, value=input_header[key], comment=_get_comment(key, spec, input_header)
                ),
                end=True,
            )
    unformatted_keys = set(input_header.keys()).difference(output_header.keys())
    if unformatted_keys:
        raise ValueError(
            "Some keys have not been formatted, which means"
            f" they are not present in the specification:\n {unformatted_keys}"
        )
    return output_header



from typing import Any

import pytest

from dkist_fits_specifications.spec214 import (
    define_214_schema_expansions,
    load_full_spec214,
    load_processed_spec214,
    load_spec214,
)
from dkist_fits_specifications.utils.frozendict import frozendict


def test_load_full_214():
    spec = load_full_spec214()
    visp = spec["visp"]
    assert "VSPNUMST" not in visp
    assert "VSPSTNUM" not in visp
    assert "IPTASK" not in spec["dkist-op"]
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], frozendict)
    assert isinstance(spec["fits"]["NAXIS"], frozendict)

    # No sections should be empty
    for key in spec:
        assert spec[key], f"The {key} section is empty"


def test_load_214():
    spec = load_spec214()
    assert isinstance(spec, frozendict)
    assert isinstance(spec["fits"], frozendict)
    assert isinstance(spec["fits"]["NAXIS"], frozendict)


def test_expanded_schema():
    schemas = load_processed_spec214(
        DAAXES=2,
        DEAXES=1,
        NAXIS=3,
        DNAXIS=5,
        ZIMAGE=True,
        ZVAL1=1,
        ZVAL2=2,
        ZVAL3=3,
        TFIELDS=5,
        NPROPOS=2,
        NEXPERS=5,
        INSTRUME="notthedkist",
    )
    assert "DINDEX3" in schemas["dataset"]
    assert "NAXIS1" in schemas["fits"]
    assert "DTYPE5" in schemas["dataset"]
    assert "EXPRID05" in schemas["dkist-id"]
    assert "PROPID02" in schemas["dkist-id"]
    for percentile in [1, 10, 25, 75, 90, 95, 98, 99]:
        assert f"DATAP{str(percentile).zfill(2)}" in schemas["stats"]
    assert "ZNAME3" in schemas["compression"]
    assert "TFORM4" in schemas["compression"]
    assert "CRPIX3" in schemas["telescope"]
    assert "CRPIX3A" in schemas["telescope"]
    assert "NBIN3" in schemas["camera"]
    for i in range(1, 4):
        for j in range(1, 4):
            assert f"PC{i}_{j}" in schemas["telescope"]


def test_define_214_schema_expansion_duplication():
    """
    Given: the list of requested spec 214 expansions
    When: checking the indices for each expansion
    Then: None of them match (all expansions are unique)
    """
    expansions = define_214_schema_expansions(
        DAAXES=2,
        DEAXES=1,
        NAXIS=3,
        DNAXIS=5,
        ZIMAGE=True,
        ZVAL1=1,
        ZVAL2=2,
        ZVAL3=3,
        TFIELDS=5,
        NPROPOS=2,
        NEXPERS=5,
        INSTRUME="notthedkist",
    )
    expansion_indices = [e.index for e in expansions]
    assert len(expansion_indices) == len(set(expansion_indices))


@pytest.mark.parametrize("instrument", ["cryo-nirsp", "dl-nirsp", "vbi", "visp", "vtf"])
def test_instrument_required(instrument):
    """
    Given: An instrument table yaml and a header with INSTRUME set to that instrument
    When: Building the 214 schema
    Then: All keys from the instrument table are set to "required"
    """
    glob_name = instrument.lower().replace(
        "-", ""
    )  # For cryo and dl b/c the yamls don't have dashes

    instrument_schema = load_processed_spec214(glob=glob_name, INSTRUME=instrument)[glob_name]
    for spec_fields in instrument_schema.values():
        assert spec_fields.get("required", False)


@pytest.mark.parametrize("instrument", ["cryo-nirsp", "dl-nirsp", "vbi", "visp", "vtf"])
def test_instrument_not_required(instrument):
    """
    Given: An instrument table yaml and a header with INSTRUME set to NOT that instrument
    When: Building the 214 schema
    Then: All keys from the instrument table are NOT set to "required"
    """
    glob_name = instrument.lower().replace(
        "-", ""
    )  # For cryo and dl b/c the yamls don't have dashes

    instrument_schema = load_processed_spec214(glob=glob_name, INSTRUME="notthedkist")[glob_name]
    for spec_key_fields in instrument_schema.values():
        assert not spec_key_fields.get("required", True)


@pytest.fixture(scope="session")
def cryo_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "cryo-nirsp", "CNMODNST": 2, "CNSPINMD": "Stepped"}


@pytest.fixture(scope="session")
def dl_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "dl-nirsp", "DLPOLMD": "Full Stokes"}


@pytest.fixture(scope="session")
def visp_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "visp", "VSPPOLMD": "observe_polarimetric"}


@pytest.mark.parametrize(
    "header_fixture",
    [
        pytest.param("cryo_polarimetric_headers", id="cryo-nirsp"),
        pytest.param("dl_polarimetric_headers", id="dl-nirsp"),
        pytest.param("visp_polarimetric_headers", id="visp"),
    ],
)
def test_polarimetric_required(header_fixture, request):
    """
    Given: The 214 tables and a header corresponding to a polarimetric dataset
    When: Building the 214 schema
    Then: All keys marked `polarimetric_required` have `required` set to True
    """
    header_vals = request.getfixturevalue(header_fixture)
    pol_schema = load_processed_spec214(glob="pac", **header_vals)["pac"]

    target_keys = ["POL_NOIS", "POL_SENS", "STOKES"]
    for spec_key in target_keys:
        assert pol_schema[spec_key].get("required", False)


@pytest.fixture(scope="session")
def cryo_non_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "cryo-nirsp", "CNMODNST": 1, "CNSPINMD": "Foo"}


@pytest.fixture(scope="session")
def dl_non_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "dl-nirsp", "DLPOLMD": "Stokes-I"}


@pytest.fixture(scope="session")
def visp_non_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "visp", "VSPPOLMD": "observe_intensity"}


@pytest.fixture(scope="session")
def vbi_non_polarimetric_headers() -> dict[str, Any]:
    return {"INSTRUME": "vbi"}


@pytest.mark.parametrize(
    "header_fixture",
    [
        pytest.param("cryo_non_polarimetric_headers", id="cryo-nirsp"),
        pytest.param("dl_non_polarimetric_headers", id="dl-nirsp"),
        pytest.param("visp_non_polarimetric_headers", id="visp"),
        pytest.param("vbi_non_polarimetric_headers", id="vbi"),
    ],
)
def test_polarimetric_not_required(header_fixture, request):
    """
    Given: The 214 tables and a header corresponding to a NON-polarimetric dataset
    When: Building the 214 schema
    Then: All keys marked `polarimetric_required` have `required` set to False
    """
    header_vals = request.getfixturevalue(header_fixture)
    pol_schema = load_processed_spec214(glob="pac", **header_vals)["pac"]

    target_keys = ["POL_NOIS", "POL_SENS", "STOKES"]
    for spec_key in target_keys:
        assert not pol_schema[spec_key].get("required", True)


"""def test_spec_122_section():
    schemas = load_expanded_spec214(DAAXES=2, DEAXES=1, NAXIS=2, DNAXIS=5, INSTRUME="notthedkist")
    assert 'copy122' in schemas
    assert 'DATE-OBS' in schemas['copy122']"""



from dkist_fits_specifications.spec214.level0 import (
    load_level0_spec214,
    spec214_122_key_map,
)


def test_key_map():
    key_map = spec214_122_key_map()
    assert key_map["LINEWAV"] == "WAVELNTH"


def test_level0_spec():
    spec = load_level0_spec214()
    assert "WAVELNTH" in spec["fits"]
    assert spec["dataset"]["LINEWAV"]["rename"] == "WAVELNTH"

    assert "VSPNUMST" in spec["visp"]
    assert "VSPSTNUM" in spec["visp"]
    assert "IPTASK" in spec["dkist-op"]



"""
Functions and schemas relating to specification 214, for level 0 FITS files.

This submodule generates a hybrid spec 122 / spec 214 schema for level 0 files
ingested by the data center.
"""
from typing import Optional
from functools import cache

from dkist_fits_specifications import spec122, spec214
from dkist_fits_specifications.utils import (
    frozendict,
    raw_schema_type_hint,
    schema_type_hint,
)

__all__ = ["load_level0_spec214"]


def rename_only_spec214(schema: raw_schema_type_hint) -> schema_type_hint:
    """
    Parse a 214 schema and filter out all keys which do not have a type of rename.
    """
    new_schema = {}
    for key, key_schema in schema[1].items():
        if "rename" in key_schema:
            new_schema[key] = key_schema

    return spec214.preprocess_schema((schema[0], new_schema), include_level0=True)


def spec214_key_lookup(
    spec_122_schemas: dict[str, schema_type_hint] = None,
    raw_214_schemas: dict[str, raw_schema_type_hint] = None,
) -> dict[str, tuple[str, str, str]]:
    """
    Generate a mapping of the level 0 214 name to a tuple of (122_key, 122_section, 214_section)
    """
    # If custom schemas are not passed, load the full schemas
    if spec_122_schemas is None:
        spec_122_schemas = spec122.load_spec122()
    if raw_214_schemas is None:
        raw_214_schemas = spec214.load_raw_spec214()

    spec_214_122_map = {}

    for section_214, schema_214 in raw_214_schemas.items():
        combined_schema = {
            name: {**schema_214[0]["spec214"], **schema} for name, schema in schema_214[1].items()
        }
        for key_214, key_schema in combined_schema.items():
            name_122 = key_schema.get("rename")
            if name_122 is not None:
                section_122 = key_schema["section_122"]
                spec_214_122_map[key_214] = (name_122, section_122, section_214)

    return spec_214_122_map


def spec214_122_key_map(
    spec_122_schemas: dict[str, schema_type_hint] = None,
    raw_214_schemas: dict[str, raw_schema_type_hint] = None,
) -> dict[str, str]:
    """
    Generate a mapping of level0 214 key to 122 for translation of a 122 header.
    """
    return {
        key: value[0]
        for key, value in spec214_key_lookup(spec_122_schemas, raw_214_schemas).items()
    }


@cache
def load_level0_spec214(glob: Optional[str] = None) -> dict[str, schema_type_hint]:
    """
    Return the loaded schemas for Level 0 214 files.

    This schema is a superset of the 122 schemas, with all 214 key schemas
    added if the 214 key is a rename of a 122 key.
    """

    spec_122_schemas = spec122.load_spec122(glob)
    raw_214_schemas = spec214.load_raw_spec214(glob)

    spec_214_122_map = spec214_key_lookup(spec_122_schemas, raw_214_schemas)

    level0_schemas = {}

    for key_214, (key_122, section_122, section_214) in spec_214_122_map.items():
        rename_214 = rename_only_spec214(raw_214_schemas[section_214])
        schema_122 = spec_122_schemas[section_122]
        level0_schemas[section_214] = {**schema_122, **rename_214}

    return frozendict(level0_schemas)



#!/usr/bin/env python

import os

from setuptools import setup

VERSION_TEMPLATE = """
# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    __version__ = get_version(root='..', relative_to=__file__)
except Exception:
    __version__ = '{version}'
""".lstrip()

setup(
    use_scm_version={
        "write_to": os.path.join("dkist_fits_specifications", "version.py"),
        "write_to_template": VERSION_TEMPLATE,
    },
)



"""
Functions and variables common to all specifications.
"""
from typing import Any, Tuple, Optional, cast
from pathlib import Path
from functools import cache

import yaml

from dkist_fits_specifications.utils.frozendict import frozendict

__all__ = [
    "load_raw_spec",
    "raw_schema_type_hint",
    "schema_type_hint",
]

schema_type_hint = frozendict[str, frozendict[str, Any]]
"""
A type hint for a single validation schema.

As returned by `.load_spec122` and `.load_spec214`.
"""

# For some reason using tuple and not typing.Tuple here causes the doc build to fail
raw_schema_type_hint = Tuple[frozendict[str, Any], schema_type_hint]
"""
A Type hint for a single raw schema, as loaded directly from a yaml file.

This type is returned by the `.load_raw_spec122` and `.load_raw_spec214` methods.
"""


@cache
def load_raw_spec(
    base_path: Path, glob: Optional[str] = None
) -> frozendict[str, raw_schema_type_hint]:
    """
    Load raw schemas from the yaml files in ``base_path``.

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    Returns
    -------
    raw_schemas
        The schemas as loaded from the yaml files.
    """
    if glob is None:
        glob = "*"

    files = Path(base_path).glob(f"{glob}.yml")

    raw_schemas = {}
    for fname in files:
        schema_name = fname.stem
        with open(fname, encoding="utf8") as fobj:
            raw_schema = tuple(yaml.load_all(fobj, Loader=yaml.SafeLoader))

            # Because this function is cached, we want to strongly discourage
            # modification of the return values. We do this by wrapping the
            # expected tree of dicts into frozendict objects which disallow
            # modification
            frozen_header = {}
            for key, head in raw_schema[0].items():
                frozen_header[key] = frozendict(head)
            frozen_key_schemas = {}
            for key, schema in raw_schema[1].items():
                frozen_key_schemas[key] = frozendict(schema)
            raw_schema = (frozendict(frozen_header), frozendict(frozen_key_schemas))

        # Apply a more specific type hint to the loaded schema
        raw_schemas[schema_name] = cast(raw_schema_type_hint, raw_schema)

    return frozendict(raw_schemas)



import pytest

from dkist_fits_specifications.utils import frozendict
from dkist_fits_specifications.utils.spec_processors import (
    ExpansionIndex,
    expand_schema,
)


def test_expansion_index():
    """
    Given: a schema and a collection of ExpansionIndex instances
    When: when applying the expansions to the schema
    Then: expanded keys are replaced
    """
    # Single index size one expansion
    k_expansion = ExpansionIndex(index="k", size=1, values=range(1, 6))
    # Double index size two expansion
    nn_expansion = ExpansionIndex(index="nn", size=2, values=range(3, 6))
    # Single index larger size expansion
    q_expansion = ExpansionIndex(index="q", size=2, values=range(10, 16))
    # Custom range expansion
    custom_expansion_values = [1, 3, 6, 9, 13, 56]
    aa_expansion = ExpansionIndex(index="aa", size=2, values=custom_expansion_values)
    # Keys needing two expansions
    i_expansion = ExpansionIndex(index="i", size=1, values=range(1, 4))
    j_expansion = ExpansionIndex(index="j", size=1, values=range(1, 4))
    original_schema = frozendict(
        {
            "FOO<k>": {"a": 1, "b": 2},
            "OTHERKEY": {"c": 3, "d": 4},  # This key is not expanded
            "BAR<nn>": {"e": 5, "f": 6},
            "CUSTOM<aa>": {"g": 7, "h": 8},
            "PC<i>_<j>": {"i": 9, "j": 10},
            "QQQ<q>": {"k": 11, "l": 12},
        }
    )
    expanded_schema = expand_schema(
        schema=original_schema,
        expansions=[k_expansion, nn_expansion, aa_expansion, i_expansion, j_expansion, q_expansion],
    )

    for i in range(1, 6):
        assert f"FOO{i}" in expanded_schema
    for j in range(3, 6):
        assert f"BAR{str(j).zfill(2)}" in expanded_schema
    for k in custom_expansion_values:
        assert f"CUSTOM{str(k).zfill(2)}" in expanded_schema
    for q in range(10, 16):
        assert f"QQQ{q}" in expanded_schema
    assert "OTHERKEY" in expanded_schema
    for key in original_schema.keys():
        if "<" in key:
            assert key not in expanded_schema


def test_expansion_making_a_key_too_long():
    """
    Given: a schema and an expansion
    When: the expansion makes the keyword longer than allowed by FITS
    Then: an error is raised
    """
    k_expansion = ExpansionIndex(index="k", size=2, values=range(1, 12))
    schema = frozendict(
        {
            "FOOBARS<k>": {"a": 1, "b": 2},
        }
    )
    with pytest.raises(ValueError):
        _ = expand_schema(schema=schema, expansions=[k_expansion])


def test_expansion_with_values_greater_than_size():
    """
    Given: a schema and an expansion
    When: the expansion takes up more size than given by the 'size' keyword
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        _ = ExpansionIndex(index="k", size=2, values=range(1, 101))



"""
Functions and schemas relating to specification 214, for level 1 FITS files.

The 214 schemas are more complex than the 122 ones, as they have more uses, and
are tightly coupled to the 122 schemas.
There are three main variants to the 214 schemas:

* The **raw schemas** as saved in the yaml files. Each 214 raw schema can
  reference a 122 schema, from which it can copy the 'type', 'default_value',
  'values', and 'format' fields for keys. A key can be marked as ``copy:
  True`` to indicate that the schema **and** the value should be the same in
  214 as 122, or it can be marked as ``copy: 'schema'`` to indicate that just
  the schema is the same for that key, but the value will be regenerated.
* The **full schemas**. The full schemas are processed to merge the values in
  the schema header, and any values from the corresponding 122 schema. These
  schemas contain a lot more fields than the 122 schemas do, as these schemas
  are also used to generate documentation describing the specification.
* The **validation schema**. These schemas (returned by `.load_spec214`) are
  the equivalent to the schemas returned by `.load_spec122`. They have all
  the extra fields present in the "full" schemas removed.

"""
import copy
from typing import Any, Optional
from pathlib import Path
from functools import cache

import astropy.units as u
import yamale  # type: ignore

from dkist_fits_specifications import spec122, validation_schema
from dkist_fits_specifications.utils import (
    frozendict,
    load_raw_spec,
    raw_schema_type_hint,
    schema_type_hint,
)

__all__ = ["load_raw_spec214", "load_full_spec214", "load_spec214", "expand_214_schema"]

from dkist_fits_specifications.utils.spec_processors import (
    ConditionalRequirement,
    ExpansionIndex,
    InstrumentRequired,
    PolarimetricRequired,
    expand_schema,
    update_schema_requiredness,
)

BASE_PATH = Path(__file__).parent / "schemas"

full_schema_214 = yamale.make_schema(Path(__file__).parent / "spec214_full_schema.yml")


def update_with_122(
    global_section_122: Optional[str], schema: schema_type_hint
) -> schema_type_hint:
    """
    Update the raw spec 214 schemas with information from the 122 schemas.

    Parameters
    ----------
    section_122
        The section in the 122 specifications to update the schema with.
    schema
        The schema to update, should be the body of a unprocessed 214 schema.

    Returns
    -------
    schema
        The updated 214 schema.
    """
    # Update the 214 schema with the data in the corresponding 122 schema
    import_keys = {
        "type",
        "default_value",
        "values",
        "values_range",
        "example_values",
        "format",
        "comment",
    }

    for key, key_schema in schema.items():
        section_122 = key_schema.get("section_122", global_section_122)
        schema_122 = {}
        if section_122:
            schema_122 = spec122.load_raw_spec122(section_122)[section_122][1]

        # Copy can be "schema" but that's all we care about here
        to_copy = bool(key_schema.get("copy", False))
        key_122 = key_schema.get("rename", key if to_copy else None)

        if key_122:
            if key_122 not in schema_122:
                print(
                    f"{key_122} is not in the 122 schema for {section_122},"
                    " although the 214 section references it."
                )
                continue
            key_schema_122 = schema_122[key_122]
            update_122 = dict(filter(lambda i: i[0] in import_keys, key_schema_122.items()))
            schema[key] = {**key_schema, **update_122}

    return schema


def update_schema_with_header(raw_schema: tuple[dict, schema_type_hint]) -> schema_type_hint:
    """
    Merge the body of a schema with its header.

    Parameters
    ----------
    raw_schema
        The loaded version of a single yaml file.

    Returns
    -------
    schema
        The body of a schema, updated with the header.
    """
    header, raw_schemas = raw_schema
    header = dict(copy.deepcopy(header)["spec214"])
    # Remove the 122 section name
    header.pop("section_122")

    key_schemas = {}
    for key, key_schema in raw_schemas.items():
        updated_schema = {key: {**header, **key_schema}}
        # Expected takes the value of required unless overridden
        updated_schema[key]["expected"] = updated_schema[key].get(
            "expected", updated_schema[key]["required"]
        )
        yamale.validate(full_schema_214, [(updated_schema, None)])
        key_schemas.update(updated_schema)

    return key_schemas


def preprocess_schema(
    raw_schema: raw_schema_type_hint, include_level0=False
) -> frozendict[str, Any]:
    """
    Convert a single raw schema or a full 214 schema.

    Parameters
    ----------
    raw_schema
        The loaded version of a single yaml file.

    Returns
    -------
    schema
        The body of a schema, updated with the header and the relevant values
        from the 122 schema where appropriate.
    """
    global_section_122 = raw_schema[0]["spec214"].get("section_122")

    # Merge in the keys in the header with all the keys
    schema = update_schema_with_header(raw_schema)  # makes a nested copy

    merged_spec = update_with_122(global_section_122, schema)
    if not include_level0:
        for key, spec in tuple(merged_spec.items()):
            if spec.get("level0_only", False):
                merged_spec.pop(key)

    # Ensure the units in the comments are correct.
    # This performs two functions:
    # 1) Make sure all keys with units in their schema have units in the FITS comments.
    # 2) Ensure that if a unit is changed between 122 and 214 the comment is updated.
    frozenspec = {}
    for key, info in merged_spec.items():
        unit = info.get("units", None)
        comment = info.get("comment", "")
        # It seems astropy doesn't support a string representation of u.one
        if unit and unit != "dimensionless":
            unit = u.Unit(unit)
            unit_str = f"[{unit:fits}]"
            if not comment.startswith(unit_str):
                if comment.startswith("["):
                    _, comment = comment.split("]", maxsplit=1)
                comment = f"{unit_str} {comment.strip()}".strip()
                info["comment"] = comment

        frozenspec[key] = frozendict(info)

    return frozendict(frozenspec)


# No cache here as load_raw_spec is cached
def load_raw_spec214(glob: Optional[str] = None) -> frozendict[str, raw_schema_type_hint]:
    """
    Load the raw 214 schemas from the yaml files.

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    Returns
    -------
    raw_schemas
        The schemas as loaded from the yaml files.
    """
    return load_raw_spec(BASE_PATH, glob)


@cache
def load_full_spec214(glob: Optional[str] = None) -> frozendict[str, schema_type_hint]:
    """
    Return the full loaded schemas for DKIST Specification 214

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    """

    raw_schemas = load_raw_spec214(glob)

    schemas = {}
    for schema_name, raw_schema in raw_schemas.items():
        schema = preprocess_schema(raw_schema)
        yamale.validate(full_schema_214, [(schema, None)])
        schemas[schema_name] = schema

    return frozendict(schemas)


@cache
def load_spec214(glob: Optional[str] = None) -> frozendict[str, schema_type_hint]:
    """
    Load the simple schema version of 214 for validation or generation.

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.
    """

    full_schemas = load_full_spec214(glob)

    # Extract the keys from the base schema.
    allowed_keys = validation_schema.includes["key"].dict.keys()

    schemas = {}
    for schema_name, full_schema in full_schemas.items():
        schema = {}
        for key, key_schema in full_schema.items():
            filtered_schema = dict(filter(lambda i: i[0] in allowed_keys, key_schema.items()))
            schema[key] = frozendict(filtered_schema)
        # Validate the schema against the simple schema
        yamale.validate(validation_schema, [(schema, None)])
        schemas[schema_name] = frozendict(schema)

    return frozendict(schemas)


def load_processed_spec214(
    glob: Optional[str] = None, **header: dict[str, Any]
) -> dict[str, schema_type_hint]:
    """
    Load the 214 schema, processed based on a FITS header.

    This function currently expands the schema and applies conditional requiredness

    Parameters
    ----------
    glob
        A pattern to use to match a file, without the ``.yml`` file extension.
        Can be a section name like ``'wcs'``.

    header
       A FITS header to be used to expand the schema as needed.

    Notes
    -----
    If you expand the schema with a single header from a dataset, the expanded
    schema should be valid for all the files in that dataset.
    """
    schemas = load_spec214(glob)
    processed_schemas = {}
    for name, schema in schemas.items():
        schema = update_214_schema_requiredness(schema, **header)
        schema = expand_214_schema(schema, **header)

        processed_schemas[name] = schema

    return processed_schemas


def define_214_schema_expansions(**header: dict[str, Any]) -> list[ExpansionIndex]:
    """
    Collate the list of schema expansions to apply for spec 214
    """
    expansion_list = []
    # If the header is compressed we need to expand based on the image axes not the table axes
    if "DAAXES" in header.keys() and "DEAXES" in header.keys():
        k_expansion = ExpansionIndex(
            index="k",
            size=1,
            values=range(header["DAAXES"] + 1, header["DAAXES"] + header["DEAXES"] + 1),
        )
        expansion_list.append(k_expansion)
    if "DNAXIS" in header.keys():
        d_expansion = ExpansionIndex(index="d", size=1, values=range(1, header["DNAXIS"] + 1))
        expansion_list.append(d_expansion)
    if "NPROPOS" in header.keys():
        rr_expansion = ExpansionIndex(index="rr", size=2, values=range(1, header["NPROPOS"] + 1))
        expansion_list.append(rr_expansion)
    if "NEXPERS" in header.keys():
        ee_expansion = ExpansionIndex(index="ee", size=2, values=range(1, header["NEXPERS"] + 1))
        expansion_list.append(ee_expansion)
    if "NAXIS" in header.keys():
        n_expansion = ExpansionIndex(index="n", size=1, values=range(1, header["NAXIS"] + 1))
        i_expansion = ExpansionIndex(index="i", size=1, values=range(1, header["NAXIS"] + 1))
        j_expansion = ExpansionIndex(index="j", size=1, values=range(1, header["NAXIS"] + 1))
        expansion_list.extend([n_expansion, i_expansion, j_expansion])
    # Compression keys
    if "ZIMAGE" in header:
        # The z index for ZNAMEz and ZVALz do not have a key which tells you how
        # many keys you have.
        zval = [k for k in header.keys() if k.startswith("ZVAL")]
        z_max = max([int(k[-1]) for k in zval])
        z_expansion = ExpansionIndex(index="z", size=1, values=range(1, z_max + 1))
        f_expansion = ExpansionIndex(index="f", size=1, values=range(1, header["TFIELDS"] + 1))
        expansion_list.extend([z_expansion, f_expansion])
    pp_expansion = ExpansionIndex(index="pp", size=2, values=[1, 10, 25, 75, 90, 95, 98, 99])
    expansion_list.append(pp_expansion)
    if "NSPECLNS" in header:
        sl_expansion = ExpansionIndex(index="sl", size=2, values=range(1, header["NSPECLNS"] + 1))
        expansion_list.append(sl_expansion)
    return expansion_list


def expand_214_schema(
    schema: schema_type_hint, **header: dict[str, Any]
) -> dict[str, frozendict[str, Any]]:
    """
    Given a 214 header expand all known indices.
    """
    return expand_schema(schema=schema, expansions=define_214_schema_expansions(**header))


def define_214_conditional_requiredness(**header: dict[str, Any]) -> list[ConditionalRequirement]:
    """
    Collate the list of conditional requirements to apply for spec 214
    """
    requirements_list = []

    instrument_requirement = InstrumentRequired(header)
    requirements_list.append(instrument_requirement)

    polarimetric_requirement = PolarimetricRequired(header)
    requirements_list.append(polarimetric_requirement)

    return requirements_list


def update_214_schema_requiredness(
    schema: schema_type_hint, **header: dict[str, Any]
) -> dict[str, frozendict[str, Any]]:
    """
    Given a 214 header, update the "required" spec field based on presence/values of spec fields that indicate conditional requiredness.
    """
    return update_schema_requiredness(
        schema=schema, requirements=define_214_conditional_requiredness(**header)
    )
